@@ -30,9 +30,11 @@
 /* .IP "\fBadd\fI sitename queueid\fR"
 /*	Inform the \fBflush\fR(8) server that the message with the specified
 /*	queue ID is queued for the specified destination.
-/* .IP "\fBsend\fI sitename\fR"
+/* .IP "\fBsend_site\fI sitename\fR"
 /*	Request delivery of mail that is queued for the specified
 /*	destination.
+/* .IP "\fBsend_file\fI queueid\fR"
+/*	Request delivery of the specified deferred message.
 /* .IP \fBrefresh\fR
 /*	Refresh non-empty per-destination logfiles that were not read in
 /*	\fB$fast_flush_refresh_time\fR hours, by simulating
@@ -44,8 +46,6 @@
 /*	This request completes in the background.
 /* .IP \fBpurge\fR
 /*	Do a \fBrefresh\fR for all per-destination logfiles.
-/* .IP "\fBsendid\fI queueid\fR"
-/*	Request delivery of the specified deferred message.
 /* SECURITY
 /* .ad
 /* .fi
@@ -53,7 +53,8 @@
 /*	talk to the network, and it does not talk to local users.
 /*	The fast flush server can run chrooted at fixed low privilege.
 /* DIAGNOSTICS
-/*	Problems and transactions are logged to \fBsyslogd\fR(8).
+/*	Problems and transactions are logged to \fBsyslogd\fR(8)
+/*	or \fBpostlogd\fR(8).
 /* BUGS
 /*	Fast flush logfiles are truncated only after a "send"
 /*	request, not when mail is actually delivered, and therefore can
@@ -96,14 +97,15 @@
 /*	The time limit for sending or receiving information over an internal
 /*	communication channel.
 /* .IP "\fBmax_idle (100s)\fR"
-/*	The maximum amount of time that an idle Postfix daemon process
-/*	waits for the next service request before exiting.
+/*	The maximum amount of time that an idle Postfix daemon process waits
+/*	for an incoming connection before terminating voluntarily.
 /* .IP "\fBmax_use (100)\fR"
-/*	The maximal number of connection requests before a Postfix daemon
-/*	process terminates.
+/*	The maximal number of incoming connections that a Postfix daemon
+/*	process will service before terminating voluntarily.
 /* .IP "\fBparent_domain_matches_subdomains (see 'postconf -d' output)\fR"
-/*	What Postfix features match subdomains of "domain.tld" automatically,
-/*	instead of requiring an explicit ".domain.tld" pattern.
+/*	A list of Postfix features where the pattern "example.com" also
+/*	matches subdomains of example.com,
+/*	instead of requiring an explicit ".example.com" pattern.
 /* .IP "\fBprocess_id (read-only)\fR"
 /*	The process ID of a Postfix command or daemon process.
 /* .IP "\fBprocess_name (read-only)\fR"
@@ -112,9 +114,13 @@
 /*	The location of the Postfix top-level queue directory.
 /* .IP "\fBsyslog_facility (mail)\fR"
 /*	The syslog facility of Postfix logging.
-/* .IP "\fBsyslog_name (postfix)\fR"
-/*	The mail system name that is prepended to the process name in syslog
-/*	records, so that "smtpd" becomes, for example, "postfix/smtpd".
+/* .IP "\fBsyslog_name (see 'postconf -d' output)\fR"
+/*	A prefix that is prepended to the process name in syslog
+/*	records, so that, for example, "smtpd" becomes "prefix/smtpd".
+/* .PP
+/*	Available in Postfix 3.3 and later:
+/* .IP "\fBservice_name (read-only)\fR"
+/*	The master.cf service name of a Postfix daemon process.
 /* FILES
 /*	/var/spool/postfix/flush, "fast flush" logfiles.
 /* SEE ALSO
@@ -123,6 +129,7 @@
 /*	postconf(5), configuration parameters
 /*	master(5), generic daemon options
 /*	master(8), process manager
+/*	postlogd(8), Postfix logging
 /*	syslogd(8), system logging
 /* README FILES
 /* .ad
@@ -143,12 +150,18 @@
 /*	IBM T.J. Watson Research
 /*	P.O. Box 704
 /*	Yorktown Heights, NY 10598, USA
+/*
+/*	Wietse Venema
+/*	Google, Inc.
+/*	111 8th Avenue
+/*	New York, NY 10011, USA
 /*--*/
 
 /* System library. */
 
 #include <sys_defs.h>
 #include <sys/stat.h>
+#include <sys/time.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <utime.h>
@@ -169,10 +182,13 @@
 #include <scan_dir.h>
 #include <stringops.h>
 #include <safe_open.h>
+#include <warn_stat.h>
+#include <midna_domain.h>
 
 /* Global library. */
 
 #include <mail_params.h>
+#include <mail_version.h>
 #include <mail_queue.h>
 #include <mail_proto.h>
 #include <mail_flush.h>
@@ -252,6 +268,15 @@ static VSTRING *flush_site_to_path(VSTRING *path, const char *site)
     const char *ptr;
     int     ch;
 
+    /*
+     * Convert the name to ASCII, so that we don't to end up with non-ASCII
+     * names in the file system. The IDNA library functions fold case.
+     */
+#ifndef NO_EAI
+    if ((site = midna_domain_to_ascii(site)) == 0)
+	return (0);
+#endif
+
     /*
      * Allocate buffer on the fly; caller still needs to clean up.
      */
@@ -263,7 +288,7 @@ static VSTRING *flush_site_to_path(VSTRING *path, const char *site)
      */
     for (ptr = site; (ch = *(unsigned const char *) ptr) != 0; ptr++)
 	if (ISALNUM(ch))
-	    VSTRING_ADDCH(path, ch);
+	    VSTRING_ADDCH(path, tolower(ch));
 	else
 	    VSTRING_ADDCH(path, '_');
     VSTRING_TERMINATE(path);
@@ -274,13 +299,6 @@ static VSTRING *flush_site_to_path(VSTRING *path, const char *site)
     return (path);
 }
 
-/* flush_policy_ok - check logging policy */
-
-static int flush_policy_ok(const char *site)
-{
-    return (domain_list_match(flush_domains, site));
-}
-
 /* flush_add_service - append queue ID to per-site fast flush logfile */
 
 static int flush_add_service(const char *site, const char *queue_id)
@@ -295,13 +313,14 @@ static int flush_add_service(const char *site, const char *queue_id)
     /*
      * If this site is not eligible for logging, deny the request.
      */
-    if (flush_policy_ok(site) == 0)
-	return (FLUSH_STAT_DENY);
+    if (domain_list_match(flush_domains, site) == 0)
+	return (flush_domains->error ? FLUSH_STAT_FAIL : FLUSH_STAT_DENY);
 
     /*
      * Map site to path and update log.
      */
-    site_path = flush_site_to_path((VSTRING *) 0, site);
+    if ((site_path = flush_site_to_path((VSTRING *) 0, site)) == 0)
+	return (FLUSH_STAT_DENY);
     status = flush_add_path(STR(site_path), queue_id);
     vstring_free(site_path);
 
@@ -371,48 +390,29 @@ static int flush_send_service(const char *site, int how)
     /*
      * If this site is not eligible for logging, deny the request.
      */
-    if (flush_policy_ok(site) == 0)
-	return (FLUSH_STAT_DENY);
+    if (domain_list_match(flush_domains, site) == 0)
+	return (flush_domains->error ? FLUSH_STAT_FAIL : FLUSH_STAT_DENY);
 
     /*
      * Map site name to path name and flush the log.
      */
-    site_path = flush_site_to_path((VSTRING *) 0, site);
+    if ((site_path = flush_site_to_path((VSTRING *) 0, site)) == 0)
+	return (FLUSH_STAT_DENY);
     status = flush_send_path(STR(site_path), how);
     vstring_free(site_path);
 
     return (status);
 }
 
-/* flush_queue_file - flush one queue file */
+/* flush_one_file - move one queue file to incoming queue */
 
-static int flush_queue_file(const char *queue_id, VSTRING *queue_file,
-			            struct utimbuf * tbuf, int how)
+static int flush_one_file(const char *queue_id, VSTRING *queue_file,
+			          struct utimbuf * tbuf, int how)
 {
-    const char *myname = "flush_queue_file";
-    static char qmgr_flush_trigger[] = {
-	QMGR_REQ_FLUSH_DEAD,		/* flush dead site/transport cache */
-    };
+    const char *myname = "flush_one_file";
     const char *queue_name;
     const char *path;
 
-    /*
-     * With the UNTHROTTLE_BEFORE strategy, we ask the queue manager to
-     * unthrottle all transports and queues before we move a deferred queue
-     * file to the incoming queue. This minimizes a race condition where the
-     * queue manager seizes a queue file before it knows that we want to
-     * flush that message.
-     * 
-     * This reduces the race condition time window to a very small amount (the
-     * flush server does not really know when the queue manager reads its
-     * command fifo). But there is a worse race, where the queue manager
-     * moves a deferred queue file to the active queue before we have a
-     * chance to expedite its delivery.
-     */
-    if (how & UNTHROTTLE_BEFORE)
-	mail_trigger(MAIL_CLASS_PUBLIC, var_queue_service,
-		     qmgr_flush_trigger, sizeof(qmgr_flush_trigger));
-
     /*
      * Some other instance of this program may flush some logfile and may
      * just have moved this queue file to the incoming queue.
@@ -493,6 +493,9 @@ static int flush_send_path(const char *path, int how)
     VSTRING *queue_file;
     VSTREAM *log;
     struct utimbuf tbuf;
+    static char qmgr_flush_trigger[] = {
+	QMGR_REQ_FLUSH_DEAD,		/* flush dead site/transport cache */
+    };
     static char qmgr_scan_trigger[] = {
 	QMGR_REQ_SCAN_INCOMING,		/* scan incoming queue */
     };
@@ -524,6 +527,23 @@ static int flush_send_path(const char *path, int how)
     if (myflock(vstream_fileno(log), INTERNAL_LOCK, MYFLOCK_OP_EXCLUSIVE) < 0)
 	msg_fatal("%s: lock fast flush logfile %s: %m", myname, path);
 
+    /*
+     * With the UNTHROTTLE_BEFORE strategy, we ask the queue manager to
+     * unthrottle all transports and queues before we move a deferred queue
+     * file to the incoming queue. This minimizes a race condition where the
+     * queue manager seizes a queue file before it knows that we want to
+     * flush that message.
+     * 
+     * This reduces the race condition time window to a very small amount (the
+     * flush server does not really know when the queue manager reads its
+     * command fifo). But there is a worse race, where the queue manager
+     * moves a deferred queue file to the active queue before we have a
+     * chance to expedite its delivery.
+     */
+    if (how & UNTHROTTLE_BEFORE)
+	mail_trigger(MAIL_CLASS_PUBLIC, var_queue_service,
+		     qmgr_flush_trigger, sizeof(qmgr_flush_trigger));
+
     /*
      * This is the part that dominates running time: schedule the listed
      * queue files for delivery by updating their file time stamps and by
@@ -558,14 +578,14 @@ static int flush_send_path(const char *path, int how)
 			 myname, path, STR(queue_id));
 	    if (dup_filter->used <= FLUSH_DUP_FILTER_SIZE)
 		htable_enter(dup_filter, STR(queue_id), 0);
-	    count += flush_queue_file(STR(queue_id), queue_file, &tbuf, how);
+	    count += flush_one_file(STR(queue_id), queue_file, &tbuf, how);
 	} else {
 	    if (msg_verbose)
 		msg_info("%s: logfile %s: skip queue file %s as duplicate",
 			 myname, path, STR(queue_file));
 	}
     }
-    htable_free(dup_filter, (void (*) (char *)) 0);
+    htable_free(dup_filter, (void (*) (void *)) 0);
     vstring_free(queue_file);
     vstring_free(queue_id);
 
@@ -575,6 +595,11 @@ static int flush_send_path(const char *path, int how)
     if (count > 0 && ftruncate(vstream_fileno(log), (off_t) 0) < 0)
 	msg_fatal("%s: truncate fast flush logfile %s: %m", myname, path);
 
+    /*
+     * Workaround for noatime mounts. Use futimes() if available.
+     */
+    (void) utimes(VSTREAM_PATH(log), (struct timeval *) 0);
+
     /*
      * Request delivery and clean up.
      */
@@ -591,11 +616,11 @@ static int flush_send_path(const char *path, int how)
     return (FLUSH_STAT_OK);
 }
 
-/* flush_send_id - flush one queue file */
+/* flush_send_file_service - flush one queue file */
 
-static int flush_send_id(const char *queue_id)
+static int flush_send_file_service(const char *queue_id)
 {
-    const char *myname = "flush_send_id";
+    const char *myname = "flush_send_file_service";
     VSTRING *queue_file;
     struct utimbuf tbuf;
     static char qmgr_scan_trigger[] = {
@@ -613,7 +638,7 @@ static int flush_send_id(const char *queue_id)
 
     queue_file = vstring_alloc(30);
     tbuf.actime = tbuf.modtime = event_time();
-    if (flush_queue_file(queue_id, queue_file, &tbuf, UNTHROTTLE_AFTER) > 0)
+    if (flush_one_file(queue_id, queue_file, &tbuf, UNTHROTTLE_AFTER) > 0)
 	mail_trigger(MAIL_CLASS_PUBLIC, var_queue_service,
 		     qmgr_scan_trigger, sizeof(qmgr_scan_trigger));
     vstring_free(queue_file);
@@ -674,6 +699,16 @@ static int flush_request_receive(VSTREAM *client_stream, VSTRING *request)
 {
     int     count;
 
+    /*
+     * Announce the protocol. Allow lazy clients to receive this together
+     * with the server's response to the client request, but also allow
+     * greedy clients to detect a service mismatch early.
+     */
+    attr_print(client_stream, ATTR_FLAG_MORE,
+	       SEND_ATTR_STR(MAIL_ATTR_PROTO, MAIL_ATTR_PROTO_FLUSH),
+	       ATTR_TYPE_END);
+    (void) vstream_fflush(client_stream);
+
     /*
      * Kluge: choose the protocol depending on the request size.
      */
@@ -705,7 +740,7 @@ static int flush_request_receive(VSTREAM *client_stream, VSTRING *request)
     else {
 	if (attr_scan(client_stream,
 		      ATTR_FLAG_MORE | ATTR_FLAG_STRICT,
-		      ATTR_TYPE_STR, MAIL_ATTR_REQ, request,
+		      RECV_ATTR_STR(MAIL_ATTR_REQ, request),
 		      ATTR_TYPE_END) != 1) {
 	    return (-1);
 	}
@@ -747,50 +782,49 @@ static void flush_service(VSTREAM *client_stream, char *unused_service,
 	    site = vstring_alloc(10);
 	    queue_id = vstring_alloc(10);
 	    if (attr_scan(client_stream, ATTR_FLAG_STRICT,
-			  ATTR_TYPE_STR, MAIL_ATTR_SITE, site,
-			  ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
+			  RECV_ATTR_STR(MAIL_ATTR_SITE, site),
+			  RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
 			  ATTR_TYPE_END) == 2
 		&& mail_queue_id_ok(STR(queue_id)))
-		status = flush_add_service(lowercase(STR(site)), STR(queue_id));
+		status = flush_add_service(STR(site), STR(queue_id));
 	    attr_print(client_stream, ATTR_FLAG_NONE,
-		       ATTR_TYPE_INT, MAIL_ATTR_STATUS, status,
+		       SEND_ATTR_INT(MAIL_ATTR_STATUS, status),
 		       ATTR_TYPE_END);
-	} else if (STREQ(STR(request), FLUSH_REQ_SEND)) {
+	} else if (STREQ(STR(request), FLUSH_REQ_SEND_SITE)) {
 	    site = vstring_alloc(10);
 	    if (attr_scan(client_stream, ATTR_FLAG_STRICT,
-			  ATTR_TYPE_STR, MAIL_ATTR_SITE, site,
+			  RECV_ATTR_STR(MAIL_ATTR_SITE, site),
 			  ATTR_TYPE_END) == 1)
-		status = flush_send_service(lowercase(STR(site)),
-					    UNTHROTTLE_BEFORE);
+		status = flush_send_service(STR(site), UNTHROTTLE_BEFORE);
 	    attr_print(client_stream, ATTR_FLAG_NONE,
-		       ATTR_TYPE_INT, MAIL_ATTR_STATUS, status,
+		       SEND_ATTR_INT(MAIL_ATTR_STATUS, status),
 		       ATTR_TYPE_END);
-	} else if (STREQ(STR(request), FLUSH_REQ_SENDID)) {
+	} else if (STREQ(STR(request), FLUSH_REQ_SEND_FILE)) {
 	    queue_id = vstring_alloc(10);
 	    if (attr_scan(client_stream, ATTR_FLAG_STRICT,
-			  ATTR_TYPE_STR, MAIL_ATTR_QUEUEID, queue_id,
+			  RECV_ATTR_STR(MAIL_ATTR_QUEUEID, queue_id),
 			  ATTR_TYPE_END) == 1)
-		status = flush_send_id(STR(queue_id));
+		status = flush_send_file_service(STR(queue_id));
 	    attr_print(client_stream, ATTR_FLAG_NONE,
-		       ATTR_TYPE_INT, MAIL_ATTR_STATUS, status,
+		       SEND_ATTR_INT(MAIL_ATTR_STATUS, status),
 		       ATTR_TYPE_END);
 	} else if (STREQ(STR(request), FLUSH_REQ_REFRESH)
 		   || STREQ(STR(request), wakeup)) {
 	    attr_print(client_stream, ATTR_FLAG_NONE,
-		       ATTR_TYPE_INT, MAIL_ATTR_STATUS, FLUSH_STAT_OK,
+		       SEND_ATTR_INT(MAIL_ATTR_STATUS, FLUSH_STAT_OK),
 		       ATTR_TYPE_END);
 	    vstream_fflush(client_stream);
 	    (void) flush_refresh_service(var_fflush_refresh);
 	} else if (STREQ(STR(request), FLUSH_REQ_PURGE)) {
 	    attr_print(client_stream, ATTR_FLAG_NONE,
-		       ATTR_TYPE_INT, MAIL_ATTR_STATUS, FLUSH_STAT_OK,
+		       SEND_ATTR_INT(MAIL_ATTR_STATUS, FLUSH_STAT_OK),
 		       ATTR_TYPE_END);
 	    vstream_fflush(client_stream);
 	    (void) flush_refresh_service(0);
 	}
     } else
 	attr_print(client_stream, ATTR_FLAG_NONE,
-		   ATTR_TYPE_INT, MAIL_ATTR_STATUS, status,
+		   SEND_ATTR_INT(MAIL_ATTR_STATUS, status),
 		   ATTR_TYPE_END);
     vstring_free(request);
     if (site)
@@ -803,23 +837,31 @@ static void flush_service(VSTREAM *client_stream, char *unused_service,
 
 static void pre_jail_init(char *unused_name, char **unused_argv)
 {
-    flush_domains = domain_list_init(match_parent_style(VAR_FFLUSH_DOMAINS),
+    flush_domains = domain_list_init(VAR_FFLUSH_DOMAINS, MATCH_FLAG_RETURN
+				   | match_parent_style(VAR_FFLUSH_DOMAINS),
 				     var_fflush_domains);
 }
 
+MAIL_VERSION_STAMP_DECLARE;
+
 /* main - pass control to the single-threaded skeleton */
 
 int     main(int argc, char **argv)
 {
-    static CONFIG_TIME_TABLE time_table[] = {
+    static const CONFIG_TIME_TABLE time_table[] = {
 	VAR_FFLUSH_REFRESH, DEF_FFLUSH_REFRESH, &var_fflush_refresh, 1, 0,
 	VAR_FFLUSH_PURGE, DEF_FFLUSH_PURGE, &var_fflush_purge, 1, 0,
 	0,
     };
 
+    /*
+     * Fingerprint executables and core dumps.
+     */
+    MAIL_VERSION_STAMP_ALLOCATE;
+
     single_server_main(argc, argv, flush_service,
-		       MAIL_SERVER_TIME_TABLE, time_table,
-		       MAIL_SERVER_PRE_INIT, pre_jail_init,
-		       MAIL_SERVER_UNLIMITED,
+		       CA_MAIL_SERVER_TIME_TABLE(time_table),
+		       CA_MAIL_SERVER_PRE_INIT(pre_jail_init),
+		       CA_MAIL_SERVER_UNLIMITED,
 		       0);
 }
